from flask import Flask
app = Flask(__name__)

@app.route('/')
def hello_world():
    return 'Hello, World!'


# this api is needed to do healthcheck, so don't remove it
@app.route('/healthz')
def hello_world():
    return 'ok'



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import pprint
import json
import click
from prettytable import PrettyTable
from kae.utils import (
    get_appname, get_current_branch, get_remote_url, get_git_tag,
    get_specs_text, error, info, fatal, handle_console_err,
    display_pods,
)


@click.argument('appname', required=False)
@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.pass_context
def get_app(ctx, appname, raw):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    with handle_console_err():
        app = kae.get_app(appname)
    if raw:
        click.echo(str(app))
    else:
        table = PrettyTable(['name', 'type', 'git', 'created'])
        table.align['git'] = 'l'
        table.add_row([app['name'], app['type'], app['git'], app['created']])
        click.echo(table)


@click.argument('appname', required=False)
@click.pass_context
def delete_app(ctx, appname):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    with handle_console_err():
        ret = kae.delete_app(appname)

    click.echo(info(json.dumps(ret)))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def delete_app_canary(ctx, appname, cluster):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    kae.set_cluster(cluster)
    with handle_console_err():
        ret = kae.delete_app_canary(appname)

    click.echo(info(json.dumps(ret)))


@click.argument('appname', required=False)
@click.argument('rules', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def set_app_abtesting_rules(ctx, appname, rules, cluster):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    rules = json.loads(rules)
    kae.set_cluster(cluster)
    with handle_console_err():
        ret = kae.set_app_abtesting_rules(appname, rules)

    click.echo(info(json.dumps(ret)))


@click.argument('appname', required=False)
@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.pass_context
def get_app_releases(ctx, appname, raw):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    with handle_console_err():
        releases = kae.get_app_releases(appname)

    if raw:
        click.echo(str(releases))
    else:
        table = PrettyTable(['name', 'tag', 'created'])
        for r in releases:
            table.add_row([appname, r['tag'], r['created']])
        click.echo(table)


@click.argument('appname', required=False)
@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def get_app_pods(ctx, appname, cluster, raw):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    kae.set_cluster(cluster)

    with handle_console_err():
        pods = kae.get_app_pods(appname)

    if raw:
        click.echo(str(pods))
    else:
        table = PrettyTable(['name', 'status', 'ready'])
        for item in pods['items']:
            name = item['metadata']['name']
            status = item['status']['phase']
            ready = sum([1 if c_status['ready'] else 0 for c_status in item['status']['container_statuses']])
            table.add_row([name, status, ready])
        click.echo(table)


@click.argument('appname', required=False)
@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def watch_app_pods(ctx, appname, cluster, raw):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    kae.set_cluster(cluster)

    # with handle_console_err():
    watcher = kae.watch_app_pods(appname)
    display_pods(kae, watcher, appname, forever=True)

    # if raw:
    #     click.echo(str(pods))
    # else:
    #     table = PrettyTable(['name', 'status', 'ready'])
    #     for item in pods['items']:
    #         name = item['metadata']['name']
    #         status = item['status']['phase']
    #         ready = sum([1 if c_status['ready'] else 0 for c_status in item['status']['container_statuses']])
    #         table.add_row([name, status, ready])
    #     click.echo(table)

@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.pass_context
def get_release(ctx, appname, tag, raw):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    tag = get_git_tag(git_tag=tag)

    with handle_console_err():
        r = kae.get_release(appname, tag)

    if raw:
        click.echo(str(r))
    else:
        table = PrettyTable(['name', 'tag', 'created'])
        table.add_row([appname, r['tag'], r['created']])
        click.echo(table)


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.pass_context
def get_release_specs(ctx, appname, tag):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    tag = get_git_tag(git_tag=tag)

    with handle_console_err():
        release = kae.get_release(appname, tag)
    click.echo(info(release['specs_text']))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def get_secret(ctx, appname, cluster):
    kae = ctx.obj['kae_api']
    appname = get_appname(appname=appname)
    kae.set_cluster(cluster)

    with handle_console_err():
        d = kae.get_secret(appname)

    ss = pprint.pformat(d)
    click.echo(info(ss))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.option('-f', help='filename of secret')
@click.option('--literal')
@click.option('--replace', default=False, is_flag=True, help='replae the secret data')
@click.pass_context
def set_secret(ctx, appname, cluster, f, literal, replace):
    data = None
    if literal:
        try:
            data = json.loads(literal)
        except Exception:
            fatal("can't load secret from literal, pls ensure it's a valid json")
    try:
        if f:
            with open(f, "r") as fp:
                data = json.load(fp)
    except Exception:
        fatal("can't read secret data from {}".format(f))
    if data is None:
        fatal("you must specify literal or filename")
    for k, v in data.items():
        if not isinstance(v, str):
            fatal("value of secret must be string")
        if not isinstance(k, str):
            fatal("key of secret must be string")
    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)

    appname = get_appname(appname=appname)
    with handle_console_err():
        d = kae.set_secret(appname, data, replace)
    click.echo(info(str(d)))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def get_config(ctx, appname, cluster):
    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)

    appname = get_appname(appname=appname)
    with handle_console_err():
        d = kae.get_config(appname)
    if isinstance(d, dict):
        for k, v in d.items():
            click.echo(info("key:"))
            click.echo(k)
            click.echo(info("data:"))
            click.echo(v)
    else:
        click.echo(d)


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.option('--name', default="config", help='key of the config')
@click.option('-f', help='filename of secret')
@click.option('--literal')
@click.option('--replace', default=False, is_flag=True, help='replae the configmap data')
@click.pass_context
def set_config(ctx, appname, cluster, name, f, literal, replace):
    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)

    appname = get_appname(appname=appname)
    if literal:
        data = literal
    elif f:
        try:
            with open(f, "r") as fp:
                data = fp.read()
        except Exception:
            fatal("can't read config data from {}".format(f))
    cm_data = {
        name: data
    }
    with handle_console_err():
        d = kae.set_config(appname, cm_data, replace)
    ss = pprint.pformat(d)
    click.echo(info(ss))


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.argument('git', required=False)
@click.option('-f', help='filename of specs')
@click.option('--literal')
@click.option('--force', default=False, is_flag=True, help='force KAE to replace the exist release')
@click.pass_context
def register_release(ctx, appname, tag, git, f, literal, force):
    appname = get_appname(appname=appname)
    tag = get_git_tag(git_tag=tag)

    specs_text = get_specs_text(f, literal)
    if specs_text is None:
        errmsg = [
            "specs_text is required, please use one of the instructions to specify it.",
            "1. specify --literal or -f in coomand line",
            "2. make the current workdir in the source code dir which contains app.yaml"
        ]
        fatal('\n'.join(errmsg))
    kae = ctx.obj['kae_api']
    git = git or get_remote_url(remote=ctx.obj['remotename'])
    if not git:
        fatal("git url not found, please check repository or pass argument")
    branch = get_current_branch()
    with handle_console_err():
        kae.register_release(appname, tag, git, specs_text, branch=branch, force=force)
    click.echo(info('Register %s %s %s done.' % (appname, tag, git)))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.argument('revision', default=0)
@click.pass_context
def rollback(ctx, appname, cluster, revision):
    appname = get_appname(appname=appname)
    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)
    with handle_console_err():
        kae.rollback(appname, revision)
    click.echo(info('Rollback %s(revision %s) done.' % (appname, revision)))


@click.argument('appname', required=False)
@click.option('--cluster', default='default', help='cluster name')
@click.pass_context
def renew(ctx, appname, cluster):
    appname = get_appname(appname=appname)
    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)
    with handle_console_err():
        kae.renew(appname)
    click.echo(info('Renew %s done.' % (appname, )))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import os
import yaml
import click

from kaelib.spec import app_specs_schema
from kae.utils import (
    get_appname, get_current_branch, get_remote_url, get_git_tag,
    get_specs_text, error, info, fatal, handle_console_err,
    display_pods,
)
from kae.docker import build_image, run_container


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('-f', help='filename of specs')
@click.option('--literal')
def test(appname, tag, f, literal):
    """build test image and run test script in app.yaml"""
    repo_dir = os.getcwd()
    if f:
        repo_dir = os.path.dirname(os.path.abspath(f))

    appname = get_appname(cwd=repo_dir, appname=appname)
    tag = get_git_tag(cwd=repo_dir, git_tag=tag, required=False)
    if tag is None:
        tag = 'latest'

    specs_text = get_specs_text(f, literal)
    if specs_text is None:
        errmsg = [
            "specs_text is required, please use one of the instructions to specify it.",
            "1. specify --literal or -f in coomand line",
            "2. make the current workdir in the source code dir which contains app.yaml"
        ]
        fatal('\n'.join(errmsg))
    try:
        yaml_dict = yaml.load(specs_text)
    except yaml.YAMLError as e:
        fatal('specs text is invalid yaml {}'.format(str(e)))
    try:
        specs = app_specs_schema.load(yaml_dict).data
    except Exception as e:
        fatal('specs text is invalid: {}'.format(str(e)))
    if "test" not in specs:
        fatal("no test specified in app.yaml")
    builds = specs['test']['builds']
    if len(builds) == 0:
        builds = specs["builds"]
    for build in builds:
        image_tag = build.tag if build.tag else tag
        dockerfile = build.get('dockerfile', None)
        if dockerfile is None:
            dockerfile = os.path.join(repo_dir, "Dockerfile")
        full_image_name = "{}:{}".format(build.name, image_tag)
        build_image(full_image_name, repo_dir, None, dockerfile=dockerfile)

    default_image_name = "{}:{}".format(appname, tag)
    for entrypoint in specs['test']['entrypoints']:
        image = entrypoint.image if entrypoint.image else default_image_name
        cmd = entrypoint.script
        volumes = entrypoint.get('volumes', [])
        run_container(image, volumes, cmd)


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('-f', help='filename of specs')
@click.option('--literal')
@click.option('--test', default=False, is_flag=True, help='build test image')
def build_local(appname, tag, f, literal, test):
    """build local image """
    repo_dir = os.getcwd()
    if f:
        repo_dir = os.path.dirname(os.path.abspath(f))
    appname = get_appname(cwd=repo_dir, appname=appname)
    tag = get_git_tag(cwd=repo_dir, git_tag=tag, required=False)
    if tag is None:
        tag = 'latest'

    specs_text = get_specs_text(f, literal)
    if specs_text is None:
        errmsg = [
            "specs_text is required, please use one of the instructions to specify it.",
            "1. specify --literal or -f in coomand line",
            "2. make the current workdir in the source code dir which contains app.yaml"
        ]
        fatal('\n'.join(errmsg))
    try:
        yaml_dict = yaml.load(specs_text)
    except yaml.YAMLError as e:
        fatal('specs text is invalid yaml {}'.format(str(e)))
    try:
        specs = app_specs_schema.load(yaml_dict).data
    except Exception as e:
        fatal('specs text is invalid: {}'.format(str(e)))

    builds = specs["builds"]
    if test:
        if "test" not in specs:
            fatal("no test specified in app.yaml")
        builds = specs['test']['builds']
    if len(builds) == 0:
        fatal("no builds found")

    for build in builds:
        image_tag = build.tag if build.tag else tag
        dockerfile = build.get('dockerfile', None)
        if dockerfile is None:
            dockerfile = os.path.join(repo_dir, "Dockerfile")
        full_image_name = "{}:{}".format(build.name, image_tag)
        build_image(full_image_name, repo_dir, None, dockerfile=dockerfile)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import click
from tqdm import tqdm
from reprint import output
from prettytable import PrettyTable
from .utils import (
    error, info, get_git_tag, get_remote_url, get_specs_text,
    get_appname, get_current_branch, fatal, handle_console_err,
    merge_list, display_pods,
)

def cfg_list_to_dict(lst):
    _dict = {}
    for line in lst:
        parts = line.split(',')
        if len(parts) == 2:
            _dict['*'] = {
                'request': parts[0],
                'limit': parts[1],
            }
        elif len(parts) == 3:
            try:
                idx = int(parts[0])
                if idx in _dict:
                    fatal("duplicate index")
                _dict[idx] = {
                    'request': parts[1],
                    'limit': parts[2],
                }
            except:
                fatal("invalid index")
    return _dict


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('--block', default=False, is_flag=True, help='block when exist other build task for this app')
@click.pass_context
def build_app(ctx, appname, tag, block):
    tag = get_git_tag(git_tag=tag)
    appname = get_appname(appname=appname)

    kae = ctx.obj['kae_api']
    gen = kae.build_app(appname, tag, block)
    phase = None
    try:
        m = next(gen)
        while True:
            if m['success'] is False:
                click.echo(error(m['error']))
                raise click.Abort()
            if phase != m['phase']:
                phase = m['phase']
                click.echo(info("***** PHASE {}".format(m['phase'])))

            # if m['progress']:
            #     name = m['progress']['name']
            #     click.echo(name)
            #     with tqdm(total=m['progress']['max_count']) as pbar:
            #         pbar.update(m['progress']['cur_count'])
            #         for m in gen:
            #             if not m['progress'] or m['progress']['name'] != name:
            #                 break
            #             pbar.update(m['progress']['cur_count'])
            #     continue
            raw_data = m.get('raw_data', None)
            if raw_data is None:
                raw_data = {}
            if raw_data.get('error', None):
                click.echo(error(str(raw_data)))
                raise click.Abort()

            if phase.lower() == "pushing":
                if len(raw_data) == 1 and 'status' in raw_data:
                    click.echo(raw_data['status'])
                elif 'id' in raw_data and 'status' in raw_data:
                    # TODO: make the output like docker push
                    click.echo("{}:{}".format(raw_data['id'], raw_data['status']))
                elif 'digest' in raw_data:
                    click.echo("{}: digest: {} size: {}".format(raw_data.get('status'), raw_data['digest'], raw_data.get('size')))
                else:
                    click.echo(str(m))
            else:
                click.echo(m['msg'])
            m = next(gen)
    except StopIteration:
        pass

    if phase.lower() != "finished":
        click.echo(error("Connection is closed prematurely, build task is still running in background and KAE will send you an email after build task finished."))
        raise click.Abort()
    click.echo(info('\nBuild %s %s done.' % (appname, tag)))


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('--cpus', multiple=True, help='how many CPUs to set, format `idx,req,limit` or `req,limit`, e.g. --cpu 0,1.5,2')
@click.option('--memories', multiple=True, help='how much memory to set, format `idx,req,limit` or `req,limit` e.g. --memory 0,64M,256M')
@click.option('--replicas', type=int, help='repliocas of app, e.g. --replicas 2')
@click.option('--yaml-name', default='default', help="app yaml name")
@click.option('--cluster', default='default', help='cluster name')
@click.option('--watch', default=False, is_flag=True, help='watch pods')
@click.pass_context
def deploy_app(ctx, appname, cluster, tag, cpus, memories, replicas, yaml_name, watch):
    tag = get_git_tag(git_tag=tag)
    appname = get_appname(appname=appname)

    cpus_dict = cfg_list_to_dict(cpus)
    memories_dict = cfg_list_to_dict(memories)

    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)

    with handle_console_err():
        kae.deploy_app(appname, tag, cpus_dict, memories_dict,
                       replicas, app_yaml_name=yaml_name)
    if watch:
        watcher = kae.get_app_pods(appname, watch=True)
        display_pods(kae, watcher, appname)
    click.echo(info("deploy done.."))


@click.argument('appname', required=False)
@click.argument('tag', required=False)
@click.option('--cpus', multiple=True, help='how many CPUs to set, format `idx,req,limit` or `req,limit`, e.g. --cpu 0,1.5,2')
@click.option('--memories', multiple=True, help='how much memory to set, format `idx,req,limit` or `req,limit` e.g. --memory 0,64M,256M')
@click.option('--replicas', type=int, help='repliocas of app, e.g. --replicas 2')
@click.option('--yaml-name', default='default', help="app yaml name")
@click.option('--cluster', default='default', help='cluster name')
@click.option('--watch', default=False, is_flag=True, help='watch pods')
@click.pass_context
def deploy_app_canary(ctx, appname, cluster, tag, cpus, memories, replicas, yaml_name, watch):
    tag = get_git_tag(git_tag=tag)
    appname = get_appname(appname=appname)

    cpus_dict = cfg_list_to_dict(cpus)
    memories_dict = cfg_list_to_dict(memories)

    kae = ctx.obj['kae_api']
    kae.set_cluster(cluster)

    with handle_console_err():
        kae.deploy_app_canary(appname, tag, cpus_dict, memories_dict,
                              replicas, app_yaml_name=yaml_name)
    if watch:
        watcher = kae.get_app_pods(appname, canary=True, watch=True)
        display_pods(kae, watcher, appname, canary=True)
    click.echo(info("deploy done.."))


@click.argument('appname', required=False)
@click.option('--cpus', multiple=True, help='how many CPUs to set, format `idx,req,limit` or `req,limit`, e.g. --cpu 0,1.5,2')
@click.option('--memories', multiple=True, help='how much memory to set, format `idx,req,limit` or `req,limit` e.g. --memory 0,64M,256M')
@click.option('--replicas', default=0, type=int, help='repliocas of app, e.g. --replicas 2')
@click.option('--cluster', default='default', help='cluster name')
@click.option('--watch', default=False, is_flag=True, help='watch pods')
@click.pass_context
def scale_app(ctx, appname, cpus, memories, replicas, cluster, watch):
    appname = get_appname(appname=appname)

    if not (cpus or memories or replicas):
        fatal("you must at least sapecify one of cpu memory and replicas")
    cpus_dict = cfg_list_to_dict(cpus)
    memories_dict = cfg_list_to_dict(memories)

    kae = ctx.obj['kae_api']

    kae.set_cluster(cluster)
    with handle_console_err():
        kae.scale_app(appname, cpus_dict, memories_dict, replicas)

    if watch:
        watcher = kae.get_app_pods(appname, watch=True)
        display_pods(kae, watcher, appname)
    click.echo(info("scale done.."))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import os
import shutil
from jinja2 import Template
from pkg_resources import resource_string, resource_filename

import click
from kae.utils import (
    get_appname, get_current_branch, get_remote_url, get_git_tag,
    get_specs_text, error, info, fatal, handle_console_err,
    display_pods,
)


@click.argument('app_type', required=True)
@click.pass_context
def create_web_app(ctx, app_type):
    ty = app_type.lower()
    if ty == "flask":
        create_flask_app()
    elif ty == "tornado":
        create_tornado_app()
    elif ty == "go":
        create_go_app()
    elif ty == "node":
        create_node_app()
    else:
        click.echo("unknown app type, flask, tornado, go, node are allowed")
        raise click.Abort()


def _common():
    app_yaml_tpl = Template(resource_string(__name__, "tpl/app.yaml").decode('utf8'))

    appname = click.prompt(info('Please enter app name'), type=str)
    hostname = click.prompt(info('Please enter domain'), type=str)
    port = click.prompt(info('Please enter listen port'), type=int, default=8080)
    replicas = click.prompt(info('Please enter replicas'), type=int, default=1)

    mappings = {
        "appname": appname,
        "hostname": hostname,
        "port": port,
        "replicas": replicas,
    }
    if os.path.exists(appname):
        click.echo(error("dir {} exists".format(appname)))
        raise click.Abort()

    os.mkdir(appname)
    with open(os.path.join(appname, "app.yaml"), "w") as fp:
        fp.write(app_yaml_tpl.render(**mappings))

    shutil.copyfile(
            resource_filename(__name__, "tpl/gitlab-ci.yml"),
            os.path.join(appname, ".gitlab-ci.yml"))
    return mappings


def create_flask_app():
    mappings = _common()

    dockerfile_tpl = Template(resource_string(__name__, "tpl/flask/Dockerfile").decode('utf8'))
    gunicorn_tpl = Template(resource_string(__name__, "tpl/flask/gunicorn_config.py").decode('utf8'))

    entrypoint = click.prompt('Please enter entry point', type=str, default="app:app")
    workers = click.prompt('Please enter gunicorn workers', type=int, default=1)

    mappings.update({
        "entrypoint": entrypoint,
        "workers": workers
    })
    appname = mappings['appname']
    with open(os.path.join(appname, "Dockerfile"), "w") as fp:
        fp.write(dockerfile_tpl.render(**mappings))

    with open(os.path.join(appname, "gunicorn_config.py"), "w") as fp:
        fp.write(gunicorn_tpl.render(**mappings))

    shutil.copyfile(
        resource_filename(__name__, "tpl/flask/requirements.txt"),
        os.path.join(appname, "requirements.txt"))
    shutil.copyfile(
        resource_filename(__name__, "tpl/flask/app.py"),
        os.path.join(appname, "app.py"))


def create_tornado_app():
    mappings = _common()

    dockerfile_tpl = Template(resource_string(__name__, "tpl/tornado/Dockerfile").decode('utf8'))
    app_py_tpl = Template(resource_string(__name__, "tpl/tornado/app.py").decode('utf8'))

    appname = mappings['appname']
    with open(os.path.join(appname, "Dockerfile"), "w") as fp:
        fp.write(dockerfile_tpl.render(**mappings))

    with open(os.path.join(appname, "app.py"), "w") as fp:
        fp.write(app_py_tpl.render(**mappings))

    shutil.copyfile(
        resource_filename(__name__, "tpl/tornado/requirements.txt"),
        os.path.join(appname, "requirements.txt"))


def create_go_app():
    mappings = _common()

    dockerfile_tpl = Template(resource_string(__name__, "tpl/go/Dockerfile").decode('utf8'))
    main_go_tpl = Template(resource_string(__name__, "tpl/go/main.go").decode('utf8'))

    appname = mappings['appname']
    with open(os.path.join(appname, "main.go"), "w") as fp:
        fp.write(main_go_tpl.render(**mappings))

    with open(os.path.join(appname, "Dockerfile"), "w") as fp:
        fp.write(dockerfile_tpl.render(**mappings))


def create_node_app():
    mappings = _common()

    dockerfile_tpl = Template(resource_string(__name__, "tpl/node/Dockerfile").decode('utf8'))

    appname = mappings['appname']

    with open(os.path.join(appname, "Dockerfile"), "w") as fp:
        fp.write(dockerfile_tpl.render(**mappings))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import

from .action import (
    deploy_app,
    deploy_app_canary,
    build_app,
    scale_app,
)


from .app import (
    get_app,
    delete_app,
    delete_app_canary,
    get_app_releases,
    get_app_pods,
    get_release,
    get_release_specs,
    watch_app_pods,

    get_secret,
    set_secret,
    get_config,
    set_config,

    register_release,
    rollback,
    renew,

    set_app_abtesting_rules,
)

from .job import (
    create_job,
    list_job,
    delete_job,
    get_job_log,
)

from .test import test, build_local
from .create_app import create_web_app


commands = {
    'app:get': get_app,
    'app:release': get_app_releases,
    'app:delete': delete_app,
    'app:delete_canary': delete_app_canary,
    'app:pods': get_app_pods,
    'app:watch_pods': watch_app_pods,

    'app:set_abtesting': set_app_abtesting_rules,

    'release:get': get_release,
    'release:specs': get_release_specs,

    'secret:get': get_secret,
    'secret:set': set_secret,

    'config:get': get_config,
    'config:set': set_config,

    'app:register': register_release,
    'app:rollback': rollback,
    'app:renew': renew,

    'app:deploy': deploy_app,
    'app:deploy_canary': deploy_app_canary,
    'app:build': build_app,
    'app:scale': scale_app,

    'job:create': create_job,
    'job:list': list_job,
    'job:delete': delete_job,
    'job:log': get_job_log,

    'create-web-app': create_web_app,
    'test': test,
    'build': build_local,
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import os
import shlex
import subprocess

import click

from .utils import info, error, fatal


def _docker(args, cwd=None, env=os.environ, capture_output=False, print_stdout=True):
    """
    Wrapper of Docker client. Use subprocess instead of docker-py to
    avoid API version inconsistency problems.
    Args:
        args: Argument list to pass to Docker client.
        cwd: Current working directory to run Docker under.
        env: Environemnt variable dict to pass to Docker client.
    Returns:
        Combined output of stdout + stderr, or return code (int).
    Raises:
        None.
    """

    cmd = ['docker'] + args
    env = dict(env)

    if capture_output:
        try:
            output = subprocess.check_output(
                cmd, env=env, cwd=cwd, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as e:
            output = e.output
        return output
    else:
        retcode = subprocess.call(cmd, env=env, cwd=cwd, stderr=subprocess.STDOUT,
                                  stdout=(None if print_stdout else open('/dev/null', 'w')))
        return retcode


def build_image(name, context, build_args, dockerfile=None, use_cache=True):
    click.echo(info('building image {} ...'.format(name)))
    if use_cache:
        options = ['-t', name, ]
    else:
        options = ['-t', name, '--no-cache', ]

    if build_args:
        options = ['-t', name]
        for arg in build_args:
            key, val = arg.split('=', 1)
            if val.startswith('$'):
                val = os.environ[val[1:]]
            options.append('--build-arg')
            options.append('{}={}'.format(key, val))
    if dockerfile is not None:
        options += ['-f', dockerfile]
    docker_args = ['build'] + options + ['--network', 'host', '.']
    retcode = _docker(docker_args, cwd=context)
    if retcode != 0:
        name = None
        fatal('build failed. See errors above.')
    else:
        click.echo(info('build succeeded: {}'.format(name)))
    return name


def run_container(image, volumes, cmd_list, name=None):
    if name is None:
        docker_args = [
            'run', '--rm', '--network', 'host',
        ]
    else:
        docker_args = [
            'run', '--name', name, '--rm', '--network', 'host',
        ]
    entrypoints_args = ['--entrypoint', "sh"]
    volumes_args = []
    for volume in volumes:
        volumes_args.extend(['-v', volume])
    docker_args += entrypoints_args + volumes_args + \
                   [image, '-c',  '{}'.format(" && ".join(cmd_list))]
    retcode = _docker(docker_args)
    if retcode != 0:
        name = None
        fatal('run container failed. See errors above.')
    else:
        click.echo(info('run container succeeded: {}'.format(name)))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import json
import click
from prettytable import PrettyTable

from kae.utils import (
    get_current_branch, get_remote_url, get_git_tag,
    get_specs_text, error, info, fatal, handle_console_err
)


@click.argument('jobname', required=False)
@click.argument('git', required=False)
@click.option('-f', help='filename of specs')
@click.option('--literal')
@click.pass_context
def create_job(ctx, jobname, git, f, literal):
    specs_text = get_specs_text(f, literal)
    kae = ctx.obj['kae_api']
    git = git or get_remote_url(remote=ctx.obj['remotename'])
    if not git:
        fatal("git url not found, please check repository or pass argument")
    branch = get_current_branch()
    with handle_console_err():
        kae.create_job(jobname=jobname, git=git, specs_text=specs_text, branch=branch)
    click.echo(info('Create job done.'))


@click.option('--raw', default=False, is_flag=True, help='print the raw json')
@click.pass_context
def list_job(ctx, raw):
    kae = ctx.obj['kae_api']
    with handle_console_err():
        jobs = kae.list_job()

    if raw:
        click.echo(str(jobs))
    else:
        table = PrettyTable(['name', 'status', 'user', 'created'])
        for r in jobs:
            table.add_row([r['name'], r['status'], r['nickname'], r['created']])
        click.echo(table)


@click.argument('jobname', required=True)
@click.pass_context
def delete_job(ctx, jobname):
    kae = ctx.obj['kae_api']
    with handle_console_err():
        result = kae.delete_job(jobname)

    click.echo(str(result))


@click.argument('jobname', required=True)
@click.option('--follow', default=False, is_flag=True, help='follow the log stream')
@click.pass_context
def get_job_log(ctx, jobname, follow):
    kae = ctx.obj['kae_api']
    with handle_console_err():
        if follow is False:
            result = kae.get_job_log(jobname)
            click.echo(result)
        else:
            resp = kae.get_job_log(jobname, follow)
            for m in resp:
                click.echo(m)

    click.echo(info('log end..'))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import sys
import os
import re
import errno
from os import getenv
from contextlib import contextmanager

import click
import delegator
import json
import yaml
from click import ClickException
from reprint import output
from prettytable import PrettyTable
from kaelib import KaeAPIError

_GITLAB_CI_REMOTE_URL_PATTERN = re.compile(r'http://gitlab-ci-token:(.+)@([\.\w]+)/([-\w]+)/([-/\w]+).git')


def warn(text):
    return click.style(text, fg='yellow')


def error(text):
    return click.style(text, fg='red', bold=True)


def normal(text):
    return click.style(text, fg='white')


def info(text):
    return click.style(text, fg='green')


def debug_log(fmt, *args):
    return normal(fmt % args)


def fatal(msg):
    click.echo(error(msg))
    sys.exit(1)


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


@contextmanager
def handle_console_err():
    try:
        yield
    except KaeAPIError as e:
        if e.http_code == 404:
            fatal("resource not found: {}".format(str(e)))
        elif 400 <= e.http_code <= 409:
            fatal(e.msg)
        else:
            fatal("Internel Error: {}".format(e.msg))


def get_current_branch(cwd=None):
    """inside gitlab-ci, repo is at detached state, so you cannot get branch
    name from the current git repo, but luckily there's a environment
    variable called CI_BUILD_REF_NAME"""
    ctx = click.get_current_context()
    r = delegator.run('git rev-parse --abbrev-ref HEAD', cwd=cwd)
    if r.return_code:
        if ctx.obj['debug']:
            click.echo(debug_log('get_current_branch error: (stdout)%s, (stderr)%s', r.out, r.err))

        return ''

    branch = r.out.strip()
    if branch == 'HEAD':
        branch = getenv('CI_BUILD_REF_NAME', '')

    if ctx.obj['debug']:
        click.echo(debug_log('get_branch: %s', branch))

    return branch


def get_commit_hash(cwd=None):
    """拿cwd的最新的commit hash."""
    ctx = click.get_current_context()

    r = delegator.run('git rev-parse HEAD', cwd=cwd)
    if r.return_code:
        raise ClickException(r.err)

    commit_hash = r.out.strip()
    if ctx.obj['debug']:
        click.echo(debug_log('get_commit_hash: %s', commit_hash))
    return commit_hash


def get_git_tag(cwd=None, git_tag=None, required=True):
    if git_tag is not None:
        return git_tag
    ctx = click.get_current_context()

    r = delegator.run('git describe --exact-match --abbrev=0 --tags', cwd=cwd)
    if r.return_code:
        if required is True:
            fatal(r.err)
        else:
            click.echo(warn(r.err))
            return None

    git_tag = r.out.strip()
    if ctx.obj['debug']:
        click.echo(debug_log('get_git_tag: %s', git_tag))
    if (not git_tag) and (required is True):
        fatal('git tag not specified, check repo or pass argument to it.')
    return git_tag


def get_remote_url(cwd=None, remote='origin'):
    """拿cwd的remote的url.
    发现envoy.run的command只能是bytes, 不能是unicode.
    """
    ctx = click.get_current_context()

    r = delegator.run('git remote get-url %s' % str(remote), cwd=cwd)
    if r.return_code:
        raise ClickException(r.err)

    remote = r.out.strip()

    # 对gitlab ci需要特殊处理一下
    # 丫有个特殊的格式, 不太好支持...
    match = _GITLAB_CI_REMOTE_URL_PATTERN.match(remote)
    if match:
        host = match.group(2)
        group = match.group(3)
        project = match.group(4)
        return 'git@{host}:{group}/{project}.git'.format(host=host, group=group, project=project)

    if ctx.obj['debug']:
        click.echo(debug_log('get_remote_url: %s', remote))
    return remote


def get_appname(cwd=None, appname=None, required=True):
    if appname:
        return appname
    try:
        with open(os.path.join(cwd or os.getcwd(), 'app.yaml'), 'r') as f:
            specs = yaml.load(f)
    except IOError:
        if required:
            fatal('appname not specified, check app.yaml or pass argument to it.')
        return ''
    appname = specs.get('appname', '')

    if (required is True) and (not appname):
        fatal('appname not specified, check app.yaml or pass argument to it.')
    return appname


def read_json_file(path):
    try:
        with open(path) as f:
            return json.loads(f.read())
    except (OSError, IOError):
        return None


def read_yaml_file(path):
    try:
        with open(path) as f:
            return yaml.load(f)
    except (OSError, IOError):
        return None

def write_yaml_file(data, path):
    mkdir_p(os.path.dirname(path))
    try:
        with open(path, 'w') as f:
            yaml.dump(data, f)
    except (OSError, IOError):
        return None


def get_specs_text_from_repo(cwd=None):
    try:
        with open(os.path.join(cwd or os.getcwd(), 'app.yaml'), 'r') as f:
            return f.read()
    except IOError:
        return None


def get_specs_text(fname, literal):
    if literal:
        specs_text = literal
    elif fname:
        try:
            with open(fname, "r") as fp:
                specs_text = fp.read()
        except:
            fatal("can't read specs text from {}".format(fname))
    else:
        specs_text = get_specs_text_from_repo()
    return specs_text


def merge_list(l1, l2):
    len1 = len(l1)
    len2 = len(l2)
    min_len = min(len1, len2)
    l1[:min_len] = l2[:min_len]

    if len1 < len2:
        l1.extend(l2[len1:])
    else:
        l1[len2:] = ["" for i in range(len1 - len2)]


def display_pods(kae, watcher, appname, canary=False, forever=False):
    def get_replicas(appname):
        with handle_console_err():
            dp = kae.get_app_deployment(appname, canary)
            return dp['spec']['replicas']

    def extract_data_from_pod(pod):
        status = pod['status']['phase']
        restart_count = 0
        ready_count = 0
        ready_total = len(pod['spec']['containers'])
        c_status_list = pod['status'].get('container_statuses', None)
        if c_status_list:
            for cont_status in c_status_list:
                if cont_status['ready']:
                    ready_count += 1
                else:
                    if cont_status['state'].get('terminated', None):
                        status = cont_status['state']['terminated']['reason']
                    elif cont_status['state'].get('waiting', None):
                        status = cont_status['state']['waiting']['reason']
            if cont_status['restart_count'] > restart_count:
                restart_count = cont_status['restart_count']
        return {
            'ready_count': ready_count,
            'ready_total': ready_total,
            "ready": "{}/{}".format(ready_count, ready_total),
            "name": pod['metadata']['name'],
            'status': status,
            'restarts': restart_count,
            'ip': pod['status'].get('pod_ip', None),
            'node': pod['status'].get('host_ip', None),
        }

    def print_table(pod_map, output_list):
        ready_pods = 0
        table = PrettyTable(['name', 'status', 'ready', 'restarts', 'ip', 'node'])
        for name, pod in pod_map.items():
            data = extract_data_from_pod(pod)
            table.add_row([name, data['status'], data['ready'], data['restarts'], data['ip'], data['node']])
            if data['ready_count'] == data['ready_total']:
                ready_pods += 1
        merge_list(output_list, str(table).split('\n'))
        return ready_pods

    pod_map = {}
    replicas = get_replicas(appname)

    with output(output_type="list", initial_len=10, interval=0) as output_list:
        ready_pods = print_table(pod_map, output_list)
        if ready_pods == replicas and ready_pods == len(pod_map):
            return

        for m in watcher:
            action = m['action']
            pod = m["object"]
            name = pod['metadata']['name']
            if action == 'DELETED':
                pod_map.pop(name, None)
            else:
                pod_map[name] = pod
            # display table
            ready_pods = print_table(pod_map, output_list)
            if forever is False and ready_pods == replicas and ready_pods == len(pod_map):
                return



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""

"""
from __future__ import print_function, division, absolute_import
import sys
import logging
from os import getenv
from os.path import expanduser

import click

from kaelib import KaeAPI

from kae import __VERSION__
from kae.commands import commands
from kae.utils import read_yaml_file, write_yaml_file, error

__local_commands = ("version", "test", "create-web-app", "build")

@click.group(invoke_without_command=True)
@click.option('--config-path', default=expanduser('~/.kae/config.yaml'),
              help='config file, yaml', envvar='KAE_CONFIG_PATH')
@click.option('--remotename', default='origin', help='git remote name, default to origin', envvar='KAE_REPO_NAME')
@click.option('--debug', default=False, help='enable debug output', is_flag=True)
@click.option('-v', '--version', default=False, help='show version', is_flag=True)
@click.pass_context
def kae_commands(ctx, config_path, remotename, debug, version):
    if ctx.invoked_subcommand is None:
        if version:
            print("KAE version: {}".format(__VERSION__))
            return
        else:
            click.echo(error("only -v is valid when called without command"))
            sys.exit(-1)

    ctx.obj['debug'] = debug

    if ctx.invoked_subcommand not in __local_commands:
        config = read_yaml_file(config_path)
        if not config:
            config = {}
            config['auth_token'] = getenv('KAE_AUTH_TOKEN')
            config['kae_url'] = getenv('KAE_URL', 'https://console.gtapp.xyz')
            write_yaml_file(config, config_path)
            click.echo('config saved to {}'.format(config_path))

        if debug:
            logging.basicConfig(level=logging.DEBUG, format='[%(asctime)s] [%(process)d] [%(levelname)s] [%(filename)s @ %(lineno)s]: %(message)s', datefmt='%Y-%m-%d %H:%M:%S %z')

        if not config['auth_token']:
            raise Exception('KAE_AUTH_TOKEN not found')
        kae_api = KaeAPI(config['kae_url'].strip('/'), auth_token=config['auth_token'])
        ctx.obj['kae_api'] = kae_api
        ctx.obj['remotename'] = remotename


for command, function in commands.items():
    kae_commands.command(command)(function)


@kae_commands.command()
def version():
    print("KAE version: {}, python: {}".format(__VERSION__, sys.version))


def main():
    kae_commands(obj={})
